# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""The module is considered internal."""

from collections import OrderedDict
from typing import Dict, Iterable, List, Mapping, Optional, Sequence, Tuple

import dimod
import networkx as nx
import numpy as np
import scipy.optimize

from dimod.typing import GraphLike, Variable

from penaltymodel.exceptions import ImpossiblePenaltyModel
from penaltymodel.utils import as_graph

__all__ = []


class Index:
    """Tracks the column indices of the various values of interest"""

    def __init__(self,
                 decision: Iterable[Variable],
                 auxiliary: Iterable[Variable],
                 interactions: Iterable[Tuple[Variable, Variable]]
                 ):

        start = 2  # after gap and offset
        self._decision = dict((v, i) for i, v in enumerate(decision, start))
        start += len(self._decision)
        self._auxiliary = dict((v, i) for i, v in enumerate(auxiliary, start))
        start += len(self._auxiliary)
        self._interactions = dict(
            (frozenset(e), i) for i, e in enumerate(interactions, start))
        self._stop = start + len(self._interactions)

    def __len__(self) -> int:
        return self._stop

    def auxiliaries(self) -> Sequence[int]:
        return range(2 + len(self._decision), 2 + self.num_variables())

    def decisions(self) -> Sequence[int]:
        return range(2, 2 + len(self._decision))

    @staticmethod
    def gap() -> int:
        return 0

    def interaction(self, u: Variable, v: Variable) -> int:
        return self._interactions[frozenset((u, v))]

    def num_variables(self) -> int:
        return len(self._decision) + len((self._auxiliary))

    def make_bounds(self,
                    min_classical_gap: float,
                    linear_bound: Tuple[float, float],
                    quadratic_bound: Tuple[float, float],
                    ) -> Sequence[Tuple[Optional[float], Optional[float]]]:
        bounds = [(min_classical_gap, None), (None, None)]
        bounds.extend(linear_bound for _ in range(self.num_variables()))
        bounds.extend(quadratic_bound for _ in range(len(bounds), len(self)))
        return bounds

    @staticmethod
    def offset() -> int:
        return 1

    def variable(self, v: Variable) -> int:
        try:
            return self._decision[v]
        except KeyError:
            try:
                return self._auxiliary[v]
            except KeyError:
                raise ValueError(f"no variable {v!r}") from None

    def variables(self) -> Sequence[int]:
        return range(2, 2 + self.num_variables())


def all_possible(num_variables: int) -> np.ndarray:
    """Create an array of all possible spin configurations."""
    assert 0 <= num_variables < 1 << 8
    a = np.unpackbits(np.arange(1 << num_variables, dtype=np.uint8).reshape(1 << num_variables, 1),
                      axis=1, count=num_variables, bitorder='little').astype(np.int8)
    a *= 2
    a -= 1
    return a


def next_auxiliary(state: Tuple[int, ...]) -> Tuple[int, ...]:
    s = list(state)
    for i in range(len(s)-1, -1, -1):
        if s[i] <= 0:
            s[i] = 1
            break
        else:
            s[i] = -1
    return tuple(s)


def generate(graph_like: GraphLike,
             samples_like,
             *,
             linear_bound: Tuple[float, float] = (-2, 2),
             quadratic_bound: Tuple[float, float] = (-1, 1),
             min_classical_gap: float = 2,
             ) -> Tuple[dimod.BinaryQuadraticModel, float, Dict[Tuple[int, ...], Tuple[int, ...]]]:
    """Generate a penalty model.

    This function is considered internal, it is recommended to use
    :func:`~penaltymodel.get_penalty_model` with ``use_cache=False`` instead.
    """
    graph = as_graph(graph_like)
    samples, decision = dimod.as_samples(samples_like)

    if any(v not in graph.nodes for v in decision):
        raise ValueError("the decision variables must be a subset of the graph nodes")

    # let's make things easier for ourselves by casting the samples into -1, +1
    if (samples == 0).any():
        samples = 2*samples - 1
    if not ((samples == +1) ^ (samples == -1)).all():
        raise ValueError("given samples should be 0/1 or -1/+1")

    auxiliaries = list(graph.nodes - decision)
    num_samples = samples.shape[0]
    num_variables = len(graph.nodes)
    num_auxiliary = num_variables - len(decision)

    if isinstance(samples_like, dimod.SampleSet):
        energies = samples_like.record.energy
    else:
        energies = np.zeros(num_samples)

    # construct the table
    if len(energies):
        table = {tuple(map(int, state)): energy for state, energy in zip(samples, energies)}
    else:
        table = {}

    # todo: more correctness checks, max variables==8

    # some edge cases we can easily eliminate
    if not table or not decision:
        bqm = dimod.BinaryQuadraticModel('SPIN')
        bqm.add_linear_from((v, 0) for v in graph.nodes)
        bqm.add_quadratic_from((u, v, 0) for u, v in graph.edges)
        return bqm, float('inf'), {}

    # create an object to track the columns in the LP matrix

    indexer = Index(decision, auxiliaries, graph.edges)

    # we'll use this to track where the ground states are. Note that we could
    # avoiding needing to store this in memory with some clever indexing, but
    # let's not worry too much about that unless it becomes a problem later
    ground: Dict[Tuple[int, ...], Dict[Tuple[int, ...], int]] = dict()

    # ok, let's build our matrices for the LP

    b = np.full(1 << num_variables, max(table.values(), default=0), dtype=float)

    A = np.empty((1 << len(graph), len(indexer)), dtype=np.int8)
    A[:, indexer.variables()] = all_possible(num_variables)
    for u, v in graph.edges:
        A[:, indexer.interaction(u, v)] = A[:, indexer.variable(u)] * A[:, indexer.variable(v)]
    A[:, indexer.offset()] = 1

    # the gap and b are how we distinguish between values in the table and
    # not
    for i in range(1 << num_variables):
        decision_state = tuple(A[i, indexer.decisions()])

        if decision_state in table:
            A[i, indexer.gap()] = 0
            b[i] = table[decision_state]
            ground.setdefault(decision_state, {})[tuple(A[i, indexer.auxiliaries()])] = i
        else:
            A[i, indexer.gap()] = -1

    # for now we're just trying to find feasibility, we'll optimize at the end
    c = np.zeros(len(indexer))

    # bounds are fixed
    bounds = indexer.make_bounds(min_classical_gap, linear_bound, quadratic_bound)

    # ok, we have everything in hand to start solving!
    upper_bound = list(range(A.shape[0]))
    equality: List[int] = []
    auxiliary_configurations: Dict[Tuple[int, ...], Tuple[int, ...]] = OrderedDict()

    # WLOG, we can fix one right away
    decision_state = next(state for state in ground if state not in auxiliary_configurations)
    auxiliary_configurations[decision_state] = auxiliary_state = (-1,)*num_auxiliary
    i = ground[decision_state][auxiliary_state]
    upper_bound.remove(i)
    equality.append(i)

    while True:
        A_eq = A[equality, :]
        b_eq = b[equality]
        A_ub = -A[upper_bound, :]  # negate because we want A_ub <= b_ub
        b_ub = -b[upper_bound]

        res = scipy.optimize.linprog(c, A_ub, b_ub, A_eq, b_eq, bounds=bounds, method='highs')

        if res.success:
            if len(auxiliary_configurations) == len(table):
                break

            # fix a new state
            decision_state = next(state for state in ground if state not in auxiliary_configurations)
            auxiliary_configurations[decision_state] = auxiliary_state = (-1,)*num_auxiliary
            i = ground[decision_state][auxiliary_state]
            upper_bound.remove(i)
            equality.append(i)
        else:
            # ok, we didn't succeed. So first try changing the aux state of the
            # last set
            try:
                decision_state, auxiliary_state = auxiliary_configurations.popitem()
                upper_bound.append(equality.pop())  # put it back into inequality
                while all(s == 1 for s in auxiliary_state):
                    decision_state, auxiliary_state = auxiliary_configurations.popitem()
                    upper_bound.append(equality.pop())  # put it back into inequality
            except KeyError:
                raise ImpossiblePenaltyModel("There is no BQM that can encode the given constraint") from None

            # iterate the auxiliary state
            auxiliary_configurations[decision_state] = auxiliary_state = next_auxiliary(auxiliary_state)
            i = ground[decision_state][auxiliary_state]
            upper_bound.remove(i)
            equality.append(i)

    assert res.success

    # having found something feasible, let's do one last run, this time optimizing the gap
    c[indexer.gap()] = -1
    res_opt = scipy.optimize.linprog(c, A_ub, b_ub, A_eq, b_eq, bounds=bounds, method='highs')
    if res_opt.success:
        res = res_opt
        gap = res.x[indexer.gap()]
    elif res_opt.status == 3:
        # error code 3 is unbounded objective, which can happen for a fully
        # specified problem
        gap = float('inf')
    else:
        raise RuntimeError("something went wrong")

    # let's make the BQM!
    bqm = dimod.BinaryQuadraticModel('SPIN')
    bqm.add_linear_from((v, res.x[indexer.variable(v)]) for v in graph.nodes)
    bqm.add_quadratic_from((u, v, res.x[indexer.interaction(u, v)]) for u, v in graph.edges)
    bqm.offset = res.x[indexer.offset()]

    # return which auxiliary variables are which
    aux = dict((state, dict(zip(auxiliaries, aux))) for state, aux in auxiliary_configurations.items())

    return bqm, gap, aux



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

raise ImportError(
    "penaltymodel.cache was removed in penaltymodel 1.0.0 and is no longer maintained. "
    "That library can be installed with `pip install penaltymodel-cache`. "
    "This command will also install a previous version of penaltymodel. "
    "If there was some specific functionality in maxgap that you need, please "
    "open a github issue at https://github.com/dwavesystems/penaltymodel or "
    "https://github.com/dwavesystems/dwave-ocean-sdk."
    )



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""penaltymodel provides functionality for accessing PenaltyModel factories.

Accessing Factories
-------------------

Any factories that have been identified through the :const:`FACTORY_ENTRYPOINT` entrypoint
and installed on the python path can be accessed through the :func:`get_penalty_model`
function.

Examples:
    >>> import networkx as nx
    >>> import dimod
    >>> graph = nx.path_graph(5)
    >>> decision_variables = (0, 4)  # the ends of the path
    >>> feasible_configurations = {(-1, -1), (1, 1)}  # we want the ends of the path to agree
    >>> spec = pm.Specification(graph, decision_variables, feasible_configurations, dimod.SPIN)
    >>> widget = pm.get_penalty_model(spec)

Functions and Utilities
-----------------------
"""
import warnings

import dimod

from penaltymodel.database import PenaltyModelCache
from penaltymodel.core.classes import PenaltyModel, Specification
from penaltymodel.interface import get_penalty_model as _get_penalty_model
from penaltymodel.utils import table_to_sampleset

__all__ = ['FACTORY_ENTRYPOINT', 'CACHE_ENTRYPOINT', 'get_penalty_model', 'penaltymodel_factory',
           'iter_factories', 'iter_caches']

# endpoints kept for backwards compatibility, but they do nothing
FACTORY_ENTRYPOINT = 'penaltymodel_factory'
CACHE_ENTRYPOINT = 'penaltymodel_cache'


def get_penalty_model(specification):
    """Retrieve a PenaltyModel from one of the available factories.

    Args:
        specification (:class:`.Specification`): The specification
            for the desired PenaltyModel.

    Returns:
        :class:`.PenaltyModel`/None: A PenaltyModel as returned by
        the highest priority factory, or None if no factory could
        produce it.

    Raises:
        :exc:`ImpossiblePenaltyModel`: If the specification
            describes a penalty model that cannot be built by any
            factory.

    """
    warnings.warn(
        "penaltymodel.core.get_penalty_model() function is deprecated "
        "and will be removed in penaltymodel 2.0.0, "
        "use penaltymodel.get_penalty_model() instead, which has a different "
        "interface.",
        DeprecationWarning, stacklevel=3)

    # we used to specify them per variable, but now just do it globally
    linear_bound = (
        min((b for b, _ in specification.ising_linear_ranges.values()), default=-2),
        max((b for _, b in specification.ising_linear_ranges.values()), default=2)
        )

    quadratic_bound = (
        min((b for n in specification.ising_quadratic_ranges.values() for b, _ in n.values()), default=-1),
        max((b for n in specification.ising_quadratic_ranges.values() for _, b in n.values()), default=1)
        )

    bqm, gap = _get_penalty_model(
        table_to_sampleset(specification.feasible_configurations, specification.decision_variables, specification.vartype),
        specification.graph,
        linear_bound=linear_bound,
        quadratic_bound=quadratic_bound,
        min_classical_gap=specification.min_classical_gap,
        )

    return PenaltyModel.from_specification(
        specification,
        model=bqm.change_vartype(specification.vartype, inplace=True),
        classical_gap=gap,
        ground_energy=min(specification.feasible_configurations.values(), default=0),
        )


def penaltymodel_factory(priority):
    """Decorator to assign a `priority` attribute to the decorated function.

    Args:
        priority (int): The priority of the factory. Factories are queried
            in order of decreasing priority.

    Examples:
        Decorate penalty model factories like:

        >>> @pm.penaltymodel_factory(105)
        ... def factory_function(spec):
        ...     pass
        >>> factory_function.priority
        105

    """
    warnings.warn(
        "penaltymodel_factory() is deprecated and will be removed in penaltymodel 2.0.0",
        DeprecationWarning, stacklevel=2)

    # just do nothing
    def _entry_point(f):
        return f
    return _entry_point


def iter_factories():
    """Iterate through all factories identified by the factory entrypoint.

    Yields:
        function: A function that accepts a :class:`.Specification` and
        returns a :class:`.PenaltyModel`.

    """
    warnings.warn(
        "iter_factories() is deprecated and will be removed in penaltymodel 2.0.0, "
        "use penaltymodel.get_penalty_model() directly instead.",
        DeprecationWarning, stacklevel=2)

    # This used to iterate over installed factories, but now there is only one
    # so just yield it
    yield get_penalty_model


def iter_caches():
    """Iterator over the PenaltyModel caches.

    Yields:
        function: A function that accepts a :class:`PenaltyModel` and caches
        it.

    """
    warnings.warn(
        "iter_caches() is deprecated and will be removed in penaltymodel 2.0.0, "
        "use penaltymodel.PenaltyModelCache directly instead.",
        DeprecationWarning, stacklevel=2)

    cache = PenaltyModelCache()

    def cache_function(pm: PenaltyModel):
        sampleset = table_to_sampleset(pm.feasible_configurations, pm.decision_variables, pm.vartype)
        cache.insert_penalty_model(pm.model, sampleset, pm.classical_gap)

    # This used to iterate over installed caches, but now there is only one
    # so just yield it
    yield cache_function



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Specification
-------------
"""

from numbers import Number
import itertools

import networkx as nx

import dimod


__all__ = ['Specification']


class Specification(object):
    """Specification for a PenaltyModel.

    See :class:`.PenaltyModel` documentation for a fuller description of the
    different components. A specification can be thought of as an incomplete
    penalty model.

    Args:
        graph (:class:`networkx.Graph`/iterable[edge]):
            Defines the structure of the desired binary quadratic model. Each
            node in the graph represents a variable and each edge defines an
            interaction between two variables.
            If given as an iterable of edges, the graph will be constructed
            by adding each edge to an (initially) empty graph.

        decision_variables (iterable):
            The labels of the penalty model's decision variables. Each variable label
            in `decision_variables` must correspond to a node in `graph`.
            Should be an ordered iterable of hashable labels.

        feasible_configurations (dict[tuple[int], number]/iterable[tuple[int]]):
            The set of feasible configurations. Defines the allowed configurations
            of the decision variables allowed by the constraint.
            Each feasible configuration should be a tuple, each element of which
            must be of a value matching `vartype`. If given as a dict, the key
            is the feasible configuration and the value is the desired relative
            energy. If given as an iterable, it will be case to a dict where
            the relative energies are all 0.

        vartype (:class:`dimod.Vartype`/str/set):
            The variable type desired for the penalty model.
            Accepted input values:
            :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

        ising_linear_ranges (dict[node, [number, number]], optional, default=None):
            When the penalty model is spin-valued, specifies the allowed range
            for each of the linear biases.
            If a dict, should be of the form {v: [min, max], ...} where v is
            a variable in the desired penalty model and (min, max) defines
            the acceptable range for the linear bias associated with v.
            If None, the default will be set to {v: [-1, 1], ...} for each
            v in graph.
            A partial assignment is allowed.

        ising_quadratic_ranges (dict[node, dict[node, [number, number]], optional, default=None):
            When the penalty model is spin-valued, specifies the allowed range
            for each of the quadratic biases.
            If a dict, should be of the form {v: {u: [min, max], ...}, ...} where
            u and v are variables in the desired penalty model and u, v have an
            interaction - there is an edge between nodes u, v in `graph`. (min, max)
            the acceptable range for the quadratic bias associated with u, v.
            If None, the default will be set to
            {v: {u: [min, max], ...}, u: {v: [min, max], ...}, ...} for each
            edge u, v in graph.
            A partial assignment is allowed.

    Examples:
        >>> import networkx as nx
        >>> import dimod
        >>> graph = nx.path_graph(5)
        >>> decision_variables = (0, 4)  # the ends of the path
        >>> feasible_configurations = {(-1, -1), (1, 1)}  # we want the ends of the path to agree
        >>> vartype = dimod.Vartype.SPIN
        >>> spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype)

        If we want to make the interaction between (0, 1) ferromagnetic (negative):

        >>> ising_quadratic_ranges = {0: {1: (-1, 0)}}
        >>> spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype)

    Attributes:
        decision_variables (tuple):
            The labels of the penalty model's decision variables. Each variable label
            in `decision_variables` must correspond to a node in `graph`.

        feasible_configurations (dict[tuple[int], number]):
            The set of feasible configurations. Defines the allowed configurations
            of the decision variables allowed by the constraint. The key is the
            allowed configuration, the value is the relative energy of each
            configuration.

        graph (:class:`networkx.Graph`):
            Defines the structure of the desired binary quadratic model. Each
            node in the graph represents a variable and each edge defines an
            interaction between two variables.

        ising_linear_ranges (dict[node, [number, number], optional, default=None):
            When the penalty model is spin-valued, specifies the allowed range
            for each of the linear biases.
            A dict of the form {v: [min, max], ...} where v is
            a variable in the desired penalty model and [min, max] defines
            the acceptable range for the linear bias associated with v.

        ising_quadratic_ranges (dict[node, dict[node, [number, number]]], optional, default=None):
            When the penalty model is spin-valued, specifies the allowed range
            for each of the quadratic biases.
            A dict of the form {v: {u: [min, max], ...}, u: {v: [min, max], ...}, ...} where
            u and v are variables in the desired penalty model and u, v have an
            interaction - there is an edge between nodes u, v in `graph`.

        min_classical_gap (float):
            This is a threshold value for the classical gap. It describes the minimum energy gap
            between the highest feasible state and the lowest infeasible state. Default value is 2.

    """
    @dimod.decorators.vartype_argument('vartype')
    def __init__(self, graph, decision_variables, feasible_configurations, vartype,
                 ising_linear_ranges=None, ising_quadratic_ranges=None, min_classical_gap=2):

        #
        # graph
        #
        if not isinstance(graph, nx.Graph):
            try:
                edges = graph
                graph = nx.Graph()
                graph.add_edges_from(edges)
            except TypeError:
                raise TypeError("expected graph to be a networkx Graph or an iterable of edges")
        self.graph = graph

        #
        # decision_variables
        #
        try:
            if not isinstance(decision_variables, tuple):
                decision_variables = tuple(decision_variables)
        except TypeError:
            raise TypeError("expected decision_variables to be an iterable")
        if not all(v in graph for v in decision_variables):
            raise ValueError("some vars in decision decision_variables do not have a corresponding node in graph")
        self.decision_variables = decision_variables
        num_dv = len(decision_variables)

        #
        # feasible_configurations
        #
        try:
            if not isinstance(feasible_configurations, dict):
                feasible_configurations = {config: 0.0 for config in feasible_configurations}
            else:
                if not all(isinstance(en, Number) for en in feasible_configurations.values()):
                    raise ValueError("the energy fo each configuration should be numeric")
        except TypeError:
            raise TypeError("expected decision_variables to be an iterable")
        if not all(len(config) == num_dv for config in feasible_configurations):
            raise ValueError("the feasible configurations should all match the length of decision_variables")
        self.feasible_configurations = feasible_configurations

        #
        # energy ranges
        #
        self.ising_linear_ranges = self._check_ising_linear_ranges(ising_linear_ranges, graph)
        self.ising_quadratic_ranges = self._check_ising_quadratic_ranges(ising_quadratic_ranges, graph)

        #
        # min_classical_gap
        #
        if min_classical_gap <= 0:
            raise ValueError("min_classical_gap must be a positive number")
        self.min_classical_gap = min_classical_gap

        #
        # vartype
        #
        # check that our feasible configurations match
        seen_variable_types = set().union(*feasible_configurations)
        if not seen_variable_types.issubset(vartype.value):
            raise ValueError(("feasible_configurations type must match vartype. "
                              "feasible_configurations have values {}, "
                              "values permitted by vartype are {}.").format(seen_variable_types, vartype.value))
        self.vartype = vartype

    @staticmethod
    def _check_ising_linear_ranges(linear_ranges, graph):
        """check correctness/populate defaults for ising_linear_ranges."""
        if linear_ranges is None:
            linear_ranges = {}

        for v in graph:
            if v in linear_ranges:
                # check
                linear_ranges[v] = Specification._check_range(linear_ranges[v])
            else:
                # set default
                linear_ranges[v] = [-2, 2]

        return linear_ranges

    @staticmethod
    def _check_ising_quadratic_ranges(quad_ranges, graph):
        """check correctness/populate defaults for ising_quadratic_ranges."""
        if quad_ranges is None:
            quad_ranges = {}

        # first just populate the top level so we can rely on the structure
        for u in graph:
            if u not in quad_ranges:
                quad_ranges[u] = {}

        # next let's propgate and check what is already present
        for u, neighbors in quad_ranges.items():
            for v, rang in neighbors.items():
                # check the range
                rang = Specification._check_range(rang)

                if u in quad_ranges[v]:
                    # it's symmetric
                    if quad_ranges[u][v] != quad_ranges[v][u]:
                        raise ValueError("mismatched ranges for ising_quadratic_ranges")
                quad_ranges[v][u] = quad_ranges[u][v] = rang

        # finally fill in the missing stuff
        for u, v in graph.edges:
            if u not in quad_ranges[v]:
                quad_ranges[u][v] = quad_ranges[v][u] = [-1, 1]

        return quad_ranges

    @staticmethod
    def _check_range(range_):
        """Check that a range is in the format we expect [min, max] and return"""
        try:
            if not isinstance(range_, list):
                range_ = list(range_)
            min_, max_ = range_
        except (ValueError, TypeError):
            raise TypeError("each range in ising_linear_ranges should be a list of length 2.")
        if not isinstance(min_, Number) or not isinstance(max_, Number) or min_ > max_:
            raise ValueError(("each range in ising_linear_ranges should be a 2-tuple "
                              "(min, max) where min <= max"))
        return range_

    def __len__(self):
        return len(self.graph)

    def __eq__(self, specification):
        """Implemented equality checking. """

        # for specification, graph is considered equal if it has the same nodes
        # and edges
        return (isinstance(specification, Specification) and
                self.graph.edges == specification.graph.edges and
                self.graph.nodes == specification.graph.nodes and
                self.decision_variables == specification.decision_variables and
                self.feasible_configurations == specification.feasible_configurations)

    def __ne__(self, specification):
        return not self.__eq__(specification)

    def relabel_variables(self, mapping, inplace=True):
        """Relabel the variables and nodes according to the given mapping.

        Args:
            mapping (dict): a dict mapping the current variable/node labels
                to new ones.
            inplace (bool, optional, default=True):
                If True, the specification is updated in-place; otherwise, a new specification
                is returned.

        Returns:
            :class:`.Specification`: A Specification with the variables
            relabeled according to mapping. If copy=False returns itself,
            if copy=True returns a new Specification.

        """
        graph = self.graph
        ising_linear_ranges = self.ising_linear_ranges
        ising_quadratic_ranges = self.ising_quadratic_ranges

        try:
            old_labels = set(mapping.keys())
            new_labels = set(mapping.values())
        except TypeError:
            raise ValueError("mapping targets must be hashable objects")

        for v in new_labels:
            if v in graph and v not in old_labels:
                raise ValueError(('A variable cannot be relabeled "{}" without also relabeling '
                                  "the existing variable of the same name").format(v))

        if not inplace:
            return Specification(nx.relabel_nodes(graph, mapping, copy=True),  # also checks the mapping
                                 tuple(mapping.get(v, v) for v in self.decision_variables),
                                 self.feasible_configurations,  # does not change
                                 vartype=self.vartype,  # does not change
                                 ising_linear_ranges={mapping.get(v, v): ising_linear_ranges[v] for v in graph},
                                 ising_quadratic_ranges={mapping.get(v, v): {mapping.get(u, u): r
                                                                             for u, r in neighbors.items()}
                                                         for v, neighbors in ising_quadratic_ranges.items()})
        else:
            # now we need the ising_linear_ranges and ising_quadratic_ranges
            shared = old_labels & new_labels

            if shared:
                # in this case we need to transform to an intermediate state
                # counter will be used to generate the intermediate labels, as an easy optimization
                # we start the counter with a high number because often variables are labeled by
                # integers starting from 0
                counter = itertools.count(2 * len(self))

                old_to_intermediate = {}
                intermediate_to_new = {}

                for old, new in mapping.items():
                    if old == new:
                        # we can remove self-labels
                        continue

                    if old in new_labels or new in old_labels:

                        # try to get a new unique label
                        lbl = next(counter)
                        while lbl in new_labels or lbl in old_labels:
                            lbl = next(counter)

                        # add it to the mapping
                        old_to_intermediate[old] = lbl
                        intermediate_to_new[lbl] = new

                    else:
                        old_to_intermediate[old] = new
                        # don't need to add it to intermediate_to_new because it is a self-label

                Specification.relabel_variables(self, old_to_intermediate, inplace=True)
                Specification.relabel_variables(self, intermediate_to_new, inplace=True)
                return self

            # modifies graph in place
            nx.relabel_nodes(self.graph, mapping, copy=False)

            # this is always a new object
            self.decision_variables = tuple(mapping.get(v, v) for v in self.decision_variables)

            # we can just relabel in-place without worrying about conflict
            for v in old_labels:
                if v in mapping:
                    ising_linear_ranges[mapping[v]] = ising_linear_ranges[v]
                    del ising_linear_ranges[v]

            # need to do the deeper level first
            for neighbors in ising_quadratic_ranges.values():
                for v in list(neighbors):
                    if v in mapping:
                        neighbors[mapping[v]] = neighbors[v]
                        del neighbors[v]

            # now the top level
            for v in old_labels:
                if v in mapping:
                    ising_quadratic_ranges[mapping[v]] = ising_quadratic_ranges[v]
                    del ising_quadratic_ranges[v]

            return self



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import functools
import sqlite3
import os
import json
import struct
import tempfile
import threading

from typing import Dict, Iterator, Mapping, NamedTuple, Optional, Sequence, Tuple, Union

import dimod
import homebase
import networkx as nx
import numpy as np

from penaltymodel import __version__
from penaltymodel.exceptions import MissingPenaltyModel
from penaltymodel.typing import GraphLike
from penaltymodel.utils import as_graph

__all__ = ['PenaltyModelCache']


# developer note: we could use sqlite's adaptor's methods
# but since those changes are applied globally and since we're using pretty
# standard types (NumPy arrays and NetworkX graphs) it makes sense to
# do it "by hand" rather than risk interfering with other's code.


class PenaltyModel(NamedTuple):
    bqm: dimod.BinaryQuadraticModel
    sampleset: dimod.SampleSet
    classical_gap: float


class PenaltyModelCache(contextlib.AbstractContextManager):
    """Manage a database of penalty models.

    Penalty models are stored in an :mod:`sqlite3` database.

    This class can be used as a context manager to automatically close
    the database connection on exit.

    Args:
        database:
            The path to the database the user wishes to connect to.
            The default path will depend on the operating system, certain
            environmental variables and whether it is being run inside a
            virtual environment.
            See `homebase <https://github.com/dwavesystems/homebase>`_.
            If the special database name ':memory:' is given, then a temporary
            database is created in memory.

    """

    database_schema = \
        """
        CREATE TABLE IF NOT EXISTS graph(
            num_nodes INTEGER NOT NULL,
            num_edges INTEGER NOT NULL,
            edges TEXT NOT NULL,  -- json list of lists, should be sorted (with each edge sorted)
            id INTEGER PRIMARY KEY,
            CONSTRAINT graph UNIQUE (num_nodes, edges)
        );

        CREATE TABLE IF NOT EXISTS sampleset(
            num_variables INTEGER NOT NULL,
            num_samples INTEGER NOT NULL,
            samples TEXT NOT NULL,
            energies BLOB NOT NULL,
            id INTEGER PRIMARY KEY,
            CONSTRAINT sampleset UNIQUE (
                num_variables,
                samples,
                energies)
        );

        CREATE TABLE IF NOT EXISTS binary_quadratic_model(
            bqm_data BLOB NOT NULL,
            max_quadratic_bias REAL NOT NULL,
            min_quadratic_bias REAL NOT NULL,
            max_linear_bias REAL NOT NULL,
            min_linear_bias REAL NOT NULL,
            graph_id INTEGER NOT NULL,
            id INTEGER PRIMARY KEY,
            CONSTRAINT ising_model UNIQUE (bqm_data, graph_id),
            FOREIGN KEY (graph_id) REFERENCES graph(id) ON DELETE CASCADE
        );

        CREATE TABLE IF NOT EXISTS penalty_model(
            decision_variables TEXT NOT NULL,
            classical_gap REAL NOT NULL,
            sampleset_id INT,
            bqm_id INT,
            id INTEGER PRIMARY KEY,
            CONSTRAINT penalty_model UNIQUE (decision_variables, sampleset_id, bqm_id),
            FOREIGN KEY (sampleset_id) REFERENCES sampleset(id) ON DELETE CASCADE,
            FOREIGN KEY (bqm_id) REFERENCES binary_quadratic_model(id) ON DELETE CASCADE
        );

        CREATE VIEW IF NOT EXISTS penalty_model_view AS
        SELECT
            num_variables,
            num_samples,
            samples,
            energies,

            num_nodes,
            num_edges,
            edges,

            bqm_data,
            max_quadratic_bias,
            min_quadratic_bias,
            max_linear_bias,
            min_linear_bias,

            decision_variables,
            classical_gap,
            penalty_model.id
        FROM
            binary_quadratic_model,
            sampleset,
            graph,
            penalty_model
        WHERE
            penalty_model.bqm_id = binary_quadratic_model.id
            AND sampleset.id = penalty_model.sampleset_id
            AND graph.id = binary_quadratic_model.graph_id;

        PRAGMA foreign_keys = ON;
        """

    insert_bqm_statement = \
        """
        INSERT OR IGNORE INTO binary_quadratic_model(
            bqm_data,
            max_quadratic_bias,
            min_quadratic_bias,
            max_linear_bias,
            min_linear_bias,
            graph_id)
        SELECT
            :bqm_data,
            :max_quadratic_bias,
            :min_quadratic_bias,
            :max_linear_bias,
            :min_linear_bias,
            graph.id
        FROM graph WHERE
            num_nodes = :num_nodes AND
            num_edges = :num_edges AND
            edges = :edges;
        """

    insert_graph_statement = \
        """
        INSERT OR IGNORE INTO graph(num_nodes, num_edges, edges)
        VALUES (:num_nodes, :num_edges, :edges);
        """

    insert_penalty_model_statement = \
        """
        INSERT OR IGNORE INTO penalty_model(
            decision_variables,
            classical_gap,
            sampleset_id,
            bqm_id)
        SELECT
            :decision_variables,
            :classical_gap,
            sampleset.id,
            binary_quadratic_model.id
        FROM sampleset, binary_quadratic_model, graph
        WHERE
            graph.edges = :edges AND
            graph.num_nodes = :num_nodes AND
            binary_quadratic_model.graph_id = graph.id AND
            binary_quadratic_model.bqm_data = :bqm_data AND
            sampleset.num_variables = :num_variables AND
            sampleset.samples = :samples AND
            sampleset.energies = :energies;
        """

    insert_sampleset_statement = \
        """
        INSERT OR IGNORE INTO sampleset(
            num_variables,
            num_samples,
            samples,
            energies)
        VALUES (
            :num_variables,
            :num_samples,
            :samples,
            :energies);
        """

    database_name = f'penaltymodel_v{__version__}.db'
    database_path = homebase.user_data_dir(app_name='dwave-penaltymodel-cache',
                                           app_author='dwave-systems',
                                           create=True,
                                           )

    def __init__(self, database: Optional[Union[str, os.PathLike]] = None):
        if database is None:
            database = os.path.join(self.database_path, self.database_name)
        self.conn = conn = sqlite3.connect(database)

        # add the main schema
        conn.executescript(self.database_schema)

        # give us mapping access to values returned by .execute
        conn.row_factory = sqlite3.Row

    def __exit__(self, *args):
        # todo: make reentrant
        self.close()

    def close(self):
        """Close the database connection."""
        self.conn.close()

    @staticmethod
    def encode_graph(graph_like: Union[GraphLike, dimod.BinaryQuadraticModel]
                     ) -> Dict[str, Union[int, str]]:
        """Encode a NetworkX graph or BQM to be stored in the cache."""
        if isinstance(graph_like, dimod.BinaryQuadraticModel):
            nodes = graph_like.linear.keys()
            edges = graph_like.quadratic.keys()
        else:
            graph = as_graph(graph_like)
            nodes = graph.nodes
            edges = graph.edges

        if nodes ^ range(len(nodes)):
            raise ValueError("nodes must be index-labelled")

        return dict(
            num_nodes=len(nodes),
            num_edges=len(edges),
            edges=json.dumps(sorted(map(sorted, edges)), separators=(',', ':')),
            )

    @staticmethod
    def decode_graph(row: Dict[str, Union[int, str]]) -> nx.Graph:
        """Decode a row in the cache to a NetworkX graph."""
        graph = nx.Graph()
        graph.add_nodes_from(range(row['num_nodes']))
        graph.add_edges_from(json.loads(row['edges']))
        return graph

    def insert_graph(self, graph_like: GraphLike):
        """Insert a graph into the database.

        Args:
            graph: a NetworkX graph, an integer or a list of nodes.

        Raises:
            ValueError: If the nodes of the graph are not labelled `[0, n)`.

        """
        with self.conn as cur:
            cur.execute(self.insert_graph_statement, self.encode_graph(graph_like))

    def iter_graphs(self) -> Iterator[nx.Graph]:
        """Iterate over all of the graphs in the database.

        Yields:
            All the graphs in the database, as NetworkX graphs.

        """
        yield from map(self.decode_graph, self.conn.execute("SELECT num_nodes, edges from graph;"))

    @staticmethod
    def encode_sampleset(samples_like) -> Dict[str, Union[int, str, bytes]]:
        samples, labels = dimod.as_samples(samples_like)

        if not all(i == v for i, v in enumerate(labels)):
            # we could mess with re-ordering but for now let's just do the
            # simple thing
            raise ValueError("sample labels must be labelled [0, n)")

        num_samples, num_variables = samples.shape

        if num_variables > 32:
            raise ValueError("sample set must have 32 or fewer variables")

        if isinstance(samples_like, dimod.SampleSet):
            energies = samples_like.record.energy
        else:
            energies = np.zeros(num_samples)

        order = np.lexsort(samples.transpose(), axis=0)
        samples = samples[order, :]
        energies = energies[order]

        packed = dimod.serialization.utils.pack_samples(samples > 0).flatten().tolist()

        return dict(
            num_variables=num_variables,
            num_samples=num_samples,
            samples=json.dumps(packed, separators=(',', ':')),
            energies=struct.pack('<' + 'd' * len(energies), *energies),
            )

    @staticmethod
    def decode_sampleset(row: Dict[str, Union[int, str, bytes]]) -> dimod.SampleSet:
        num_variables = row['num_variables']

        packed = np.atleast_2d(json.loads(row['samples'])).transpose()
        samples = dimod.serialization.utils.unpack_samples(packed, num_variables, dtype=np.int8)

        # convert to SPIN
        samples = 2*samples-1

        energies = struct.unpack('<' + 'd' * (len(row['energies']) // 8), row['energies'])

        return dimod.SampleSet.from_samples(samples, vartype='SPIN', energy=energies)

    def insert_sampleset(self, samples_like):
        """Insert a sample set into the database.

        Args:
            samples_like:
                Samples to add to the database.
                'samples_like' is an extension of NumPy's array_like_.
                See :func:`dimod.as_samples`.

        Raises:
            ValueError: If the variables are not labelled `[0, n)`.

        .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

        """
        with self.conn as cur:
            cur.execute(self.insert_sampleset_statement, self.encode_sampleset(samples_like))

    def iter_samplesets(self):
        """Iterate over all of the sample sets in the database."""
        select = "SELECT num_variables, num_samples, samples, energies FROM sampleset"
        yield from map(self.decode_sampleset, self.conn.execute(select))

    @staticmethod
    def encode_bqm(bqm: dimod.BinaryQuadraticModel) -> Dict[str, Union[float, bytes]]:
        return dict(
            max_quadratic_bias=bqm.quadratic.max(),
            min_quadratic_bias=bqm.quadratic.min(),
            max_linear_bias=bqm.linear.max(),
            min_linear_bias=bqm.linear.min(),
            bqm_data=bqm.to_file().read(),
            )

    @staticmethod
    def decode_bqm(row: Dict[str, Union[bytes, str, int]]) -> dimod.BinaryQuadraticModel:
        return dimod.BinaryQuadraticModel.from_file(row['bqm_data'])

    def insert_binary_quadratic_model(self, bqm: dimod.BinaryQuadraticModel):
        """Insert a binary quadratic model into the database.

        Args:
            bqm: A binary quadratic model.

        Raises:
            ValueError: If the variables of the binary quadratic model are not
                labelled `[0, n)`.

        """
        bqm = dimod.as_bqm(bqm, dtype=float)

        if bqm.vartype is not dimod.SPIN:
            bqm = bqm.change_vartype(dimod.SPIN, inplace=False)

        if not bqm.variables.is_range:
            if bqm.variables ^ range(bqm.num_variables):
                raise ValueError("BQM variables must be a range of integers")
            else:
                new = dimod.BinaryQuadraticModel(bqm.num_variables, dimod.SPIN)
                new.add_linear_from(bqm.linear)
                new.add_quadratic_from(bqm.quadratic)
                new.offset = bqm.offset
                bqm = new

        parameters = self.encode_graph(bqm)
        parameters.update(self.encode_bqm(bqm))

        with self.conn as cur:
            cur.execute(self.insert_graph_statement, parameters)
            cur.execute(self.insert_bqm_statement, parameters)

    def iter_binary_quadratic_models(self) -> Iterator[dimod.BinaryQuadraticModel]:
        """Iterate over all of the binary quadratic models in the database."""
        for bqm_data in self.conn.execute("SELECT bqm_data FROM binary_quadratic_model;"):
            yield self.decode_bqm(bqm_data)

    def insert_penalty_model(
            self,
            bqm: dimod.BinaryQuadraticModel,
            samples_like,
            classical_gap: float,
            ):
        """Insert a penalty model into the database.

        Args:
            bqm: A binary quadratic model.

            samples_like: Samples to add to the database.
                'samples_like' is an extension of NumPy's array_like_.
                See :func:`dimod.as_samples`.

            classical_gap: The classical gap. This is not checked for
                correctness.

        .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

        """

        samples, decision = samples_like = dimod.as_samples(samples_like)

        # do some input checking
        if not all(v in bqm.variables for v in decision):
            raise ValueError("bqm's variables must be a superset of the "
                             "samples_like's variables")

        if bqm.variables ^ range(bqm.num_variables):
            mapping = {v: i for i, v in enumerate(decision)}
            mapping.update((v, i) for i, v in enumerate(bqm.variables ^ decision, len(mapping)))

            return self.insert_penalty_model(bqm.relabel_variables(mapping, inplace=False), samples, classical_gap)

        parameters = self.encode_graph(bqm)
        parameters.update(self.encode_bqm(bqm))
        parameters.update(self.encode_sampleset(samples_like))
        parameters.update(
            decision_variables=json.dumps(decision, separators=(',', ':')),
            classical_gap=classical_gap,
            )

        with self.conn as cur:
            cur.execute(self.insert_graph_statement, parameters)
            cur.execute(self.insert_bqm_statement, parameters)
            cur.execute(self.insert_sampleset_statement, parameters)
            cur.execute(self.insert_penalty_model_statement, parameters)

    def iter_penalty_models(self) -> Iterator[PenaltyModel]:
        """Iterate over all of the penalty models in the database."""
        for row in self.conn.execute("SELECT * FROM penalty_model_view;"):
            yield PenaltyModel(
                    self.decode_bqm(row),
                    self.decode_sampleset(row),
                    row['classical_gap']
                )

    def retrieve(self,
                 samples_like,
                 graph_like,
                 *,
                 linear_bound: Tuple[float, float] = (-2, 2),
                 quadratic_bound: Tuple[float, float] = (-1, 1),
                 min_classical_gap: float = 2,
                 ) -> Tuple[dimod.BinaryQuadraticModel, float]:
        """Retrieve a penalty model from the database.

        Args:
            samples_like:
                The set of feasible states that form the ground states of the
                generated binary quadratic model.

                'samples_like' is an extension of NumPy's array_like_.
                See :func:`dimod.as_samples`.

                .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

            graph_like:
                Defines the structure of the desired binary quadratic model. Each
                node in the graph represents a variable and each edge defines an
                interaction between two variables.
                Can be given as a :class:`networkx.Graph`, a :class:`int`, or as
                a sequence of variable labels.

                If given as a sequence of labels, the structure will be
                fully-connected, with the variables labelled according to the
                sequence.

                If given as an int, the structure will be
                fully-connected with the variables labelled ``range(n)``.

                The nodes of the graph must be a superset of the labels of
                ``samples_like``.

                If not provided, defaults to a fully connected graph with nodes
                that are the variables of ``samples_like``.

            linear_bound:
                The range allowed for the linear biases of the binary quadratic
                model.

            quadratic_bound:
                The range allowed for the quadratic biases of the binary quadratic
                model.

            min_classical_gap:
                This is a threshold value for the classical gap. It describes the
                minimum energy gap between the highest feasible state and the
                lowest infeasible state.

        Returns:
            A 2-tuple of the binary quadratic model and the classical gap. Note
            that the binary quadratic model always has vartype ``'SPIN'``.

        """
        samples, labels = dimod.as_samples(samples_like)
        graph = as_graph(graph_like)

        # do some input checking
        if not all(v in graph.nodes for v in labels):
            raise ValueError("graph_like's nodes must be a superset of the "
                             "samples_like's variables")

        if graph.nodes ^ range(len(graph.nodes)):
            mapping = {v: i for i, v in enumerate(labels)}
            mapping.update((v, i) for i, v in enumerate(graph.nodes ^ labels, len(mapping)))

            bqm, gap = self.retrieve(samples, nx.relabel_nodes(graph, mapping, copy=True),
                                     linear_bound=linear_bound,
                                     quadratic_bound=quadratic_bound,
                                     min_classical_gap=min_classical_gap)

            inverse_mapping = dict((i, v) for v, i in mapping.items())
            return bqm.relabel_variables(inverse_mapping, inplace=True), gap

        parameters = self.encode_graph(graph)
        parameters.update(self.encode_sampleset(samples_like))
        parameters.update(
            decision_variables=json.dumps(labels, separators=(',', ':')),
            min_classical_gap=min_classical_gap,
            min_linear_bias=linear_bound[0],
            max_linear_bias=linear_bound[1],
            min_quadratic_bias=quadratic_bound[0],
            max_quadratic_bias=quadratic_bound[1],
            )

        cur = self.conn.cursor()
        cur.execute(
            """
            SELECT bqm_data, classical_gap FROM penalty_model_view
            WHERE
                -- graph:
                num_nodes = :num_nodes AND
                num_edges = :num_edges AND
                edges = :edges AND
                -- feasible_configurations:
                num_variables = :num_variables AND
                samples = :samples AND
                energies = :energies AND
                -- decision variables:
                decision_variables = :decision_variables AND
                -- bounds
                min_linear_bias >= :min_linear_bias AND
                max_linear_bias <= :max_linear_bias AND
                min_quadratic_bias >= :min_quadratic_bias AND
                max_quadratic_bias <= :max_quadratic_bias AND
                -- gap
                classical_gap >= :min_classical_gap
            ORDER BY classical_gap DESC;
            """,
            parameters
            )

        row = cur.fetchone()
        cur.close()

        if row is None:
            raise MissingPenaltyModel(
                "no penalty model with the given specification found in cache")

        return self.decode_bqm(row), row['classical_gap']


def patch_cache(database: Union[str, os.PathLike] = ':memory:'):
    """A function decorator that passes in a PenaltyModelCache as a new argument.

    Args:
        database: The database location to use. Defaults to use memory.

    """
    def _patch(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with PenaltyModelCache(database) as cache:
                return f(*args, cache, **kwargs)
        return wrapper
    return _patch


@contextlib.contextmanager
def isolated_cache(*args, **kwargs):
    """Temporarily isolate the cache.

    Can be used as a decorator or a context manager.

    This context manager is not reentrant.

    """
    with tempfile.TemporaryDirectory() as d:
        with threading.RLock():
            current = PenaltyModelCache.database_path
            PenaltyModelCache.database_path = d
            try:
                yield
            finally:
                PenaltyModelCache.database_path = current



from penaltymodel.core.classes import *
import penaltymodel.core.classes

import penaltymodel.core.exceptions

from penaltymodel.core.interface import *
import penaltymodel.core.interface

from penaltymodel.core.package_info import __version__
import penaltymodel.core.package_info

from dimod import Vartype, SPIN, BINARY



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Mapping, Optional, Sequence, Tuple

import dimod
import networkx as nx
import numpy as np

from dimod.typing import Variable, VartypeLike

from penaltymodel.typing import GraphLike

__all__ = ['as_graph']


def as_graph(graph_like: GraphLike) -> nx.Graph:
    """Create a NetworkX graph from a graph-like.

    Args:
        graph_like:
            A NetworkX :class:`~networkx.Graph`, a list of nodes or a number.
            If it is a list of nodes, then a complete graph with those nodes is
            returned.
            If it is a number, then the list of nodes is treated as
            ``range(n)``.

    Returns:
        A NetworkX graph. If ``graph_like`` is a NetworkX graph then it is
        returned.

    """
    return graph_like if isinstance(graph_like, nx.Graph) else nx.complete_graph(graph_like)


def table_to_sampleset(table: Mapping[Tuple[int, ...], float],
                       decision: Sequence[Variable],
                       vartype: Optional[VartypeLike] = None) -> dimod.SampleSet:
    """Convert a table of feasible configurations into a sample set."""
    if table:
        samples, energies = zip(*table.items())
    else:
        samples = np.empty((0, 0))
        energies = np.empty(0)

    samples = np.asarray(samples, dtype=np.int8)

    if vartype is None:
        vartype = dimod.BINARY if (samples == 0).any() else dimod.SPIN

    return dimod.SampleSet.from_samples((samples, decision), vartype=vartype, energy=energies)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

r"""This package implements the generation and caching of :term:`penalty model`\ s."""

import copy

from typing import Mapping, Optional, Sequence, Tuple

import dimod
import networkx as nx

from dimod.typing import Variable

from penaltymodel.database import PenaltyModelCache
from penaltymodel.exceptions import MissingPenaltyModel
from penaltymodel.generation import generate
from penaltymodel.typing import GraphLike

__all__ = ['get_penalty_model']


def get_penalty_model(samples_like,
                      graph_like: Optional[GraphLike] = None,
                      *,
                      linear_bound: Tuple[float, float] = (-2, 2),
                      quadratic_bound: Tuple[float, float] = (-1, 1),
                      min_classical_gap: float = 2,
                      use_cache: bool = True,
                      ) -> Tuple[dimod.BinaryQuadraticModel, float]:
    """Get a penalty model for a specific graph and set of target states.

    Args:
        samples_like:
            The set of feasible states that form the ground states of the
            generated binary quadratic model.

            ``samples_like`` is an extension of NumPy's array_like_.
            See :func:`dimod.as_samples`.

            .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

        graph_like:
            Defines the structure of the desired binary quadratic model. Each
            node in the graph represents a variable and each edge defines an
            interaction between two variables.
            Can be given as a :class:`networkx.Graph`, a :class:`int`, or as
            a sequence of variable labels.

            If given as a sequence of labels, the structure will be
            fully-connected, with the variables labelled according to the
            sequence.

            If given as an int, the structure will be
            fully-connected with the variables labelled ``range(n)``.

            The nodes of the graph must be a superset of the labels of
            ``samples_like``.

            If not provided, defaults to a fully connected graph with nodes
            that are the variables of ``samples_like``.

        linear_bound:
            The range allowed for the linear biases of the binary quadratic
            model.

        quadratic_bound:
            The range allowed for the quadratic biases of the binary quadratic
            model.

        min_classical_gap:
            This is a threshold value for the classical gap. It describes the
            minimum energy gap between the highest feasible state and the
            lowest infeasible state.

        use_cache:
            Whether to attempt to retrieve models from the cache. If ``False``,
            a new model will always be generated.

    Returns:
        A 2-tuple of the binary quadratic model and the classical gap. Note
        that the binary quadratic model always has vartype ``'SPIN'``.

    Raises:
        ImpossiblePenaltyModel:
            If it is not possible to construct a penalty model for the given
            structure and feasible states.

    Examples:

        >>> import dimod
        >>> import penaltymodel

        This example generates a penalty model for an AND gate.

        >>> bqm, gap = penaltymodel.get_penalty_model([[0, 0, 0],
        ...                                            [0, 1, 0],
        ...                                            [1, 0, 0],
        ...                                            [1, 1, 1]])

        We can then solve the generated binary quadratic model using a
        brute-force solver to confirm that we have the expected ground states.

        >>> print(dimod.ExactSolver().sample(bqm))
           0  1  2 energy num_oc.
        0 -1 -1 -1    0.0       1
        1 +1 -1 -1    0.0       1
        3 -1 +1 -1    0.0       1
        5 +1 +1 +1    0.0       1
        2 +1 +1 -1    2.0       1
        4 -1 +1 +1    2.0       1
        6 +1 -1 +1    2.0       1
        7 -1 -1 +1    6.0       1
        ['SPIN', 8 rows, 8 samples, 3 variables]

    """

    # by default, just make a compelte graph from the samples
    if graph_like is None:
        samples_like = samples, labels = dimod.as_samples(samples_like)
        graph_like = nx.complete_graph(labels)

    if use_cache:
        with PenaltyModelCache() as cache:
            try:
                return cache.retrieve(samples_like=samples_like,
                                      graph_like=graph_like,
                                      linear_bound=linear_bound,
                                      quadratic_bound=quadratic_bound,
                                      min_classical_gap=min_classical_gap,
                                      )
            except MissingPenaltyModel:
                pass  # generate

    bqm, gap, _ = generate(graph_like=graph_like,
                           samples_like=samples_like,
                           linear_bound=linear_bound,
                           quadratic_bound=quadratic_bound,
                           min_classical_gap=min_classical_gap,
                           )

    if use_cache:
        with PenaltyModelCache() as cache:
            cache.insert_penalty_model(bqm, samples_like, gap)

    return bqm, gap
